# pkt_crafter.py
from scapy.all import *
import threading

try:
    from FreewayTools.colors import cprint, wprint, cinput, iprint, ColorCodes

except ModuleNotFoundError:
    from colors import cprint, wprint, cinput, iprint, ColorCodes

packet_types = [
    Dot11, Dot11Beacon, Dot11ProbeReq, Dot11ProbeResp, Dot11AssoReq, Dot11AssoResp,
    Dot11Auth, Dot11Deauth, Dot11Disas, Dot11QoS, Dot11Elt, EAPOL
]

packet_names = sorted(set(pt.__name__ for pt in packet_types))

# Packets not included in Scapy package
packet_names.extend(["Dot11RTS", "Dot11CTS", "Dot11Ack", "Dot11CFEnd"])

packet_names_str = ', '.join(packet_names)

def random_mac():
    mac = [random.randint(0x00, 0xff) for _ in range(6)]
    mac[0] = (mac[0] & 0xfe) | 0x02
    return ':'.join(map(lambda x: format(x, '02x'), mac))

def list_packets():
    print(packet_names_str)

class CraftingTable:
    def __init__(self, interface, to_craft, addr1, addr2, addr3, ssid=None, threads=1, count=1, interval=0.1, loop=False):
        self.interface = interface
        self.to_craft = to_craft # packet_name
        self.addr1 = addr1 if addr1 != "" else random_mac()
        self.addr2 = addr2 if addr2 != "" else random_mac()
        self.addr3 = addr3 if addr3 != "" else random_mac()
        self.ssid = ssid
        self.threads = threads
        self.count = count
        self.interval = interval
        self.loop = loop
        self.global_event = threading.Event()

    def craft_packet(self):
        iprint(f"Crafting {self.to_craft} packet...")
        if self.to_craft not in packet_names:
            wprint(f"Packet type {self.to_craft} not supported.")
            return None

        packet = Dot11(addr1=self.addr1, addr2=self.addr2, addr3=self.addr3)
        radiotap = RadioTap()

        if self.to_craft == 'Dot11Beacon':
            iprint("This packet type requires only two addresses to be specified, overriding addr3 to addr2...")
            broadcast = cinput("I assume you want to broadcast (visible to all devices) (enter/n)")
            self.addr1 = self.addr1 if broadcast == "n" else "FF:FF:FF:FF:FF:FF"
            cap = cinput("Encryption code (default: ESS+privacy)") or "ESS+privacy"
            rates = cinput("Enter byte sequence of supported rates (default: \\x82\\x84\\x8b\\x96\\x0c\\x12\\x18\\x24)") or b'\x82\x84\x8b\x96\x0c\x12\x18\x24'
            packet = radiotap / Dot11(type=0, subtype=8, addr1=self.addr1, addr2=self.addr2, addr3=self.addr2) / Dot11Beacon(cap=cap)
            packet /= Dot11Elt(ID="SSID", info=self.ssid)
            packet /= Dot11Elt(ID="Rates", info=rates)
            packet /= Dot11Elt(ID="DSset", info=chr(random.randint(1, 14)))

            # RSNinfo is necessary for the beacon to appear as valid
            packet /= Dot11Elt(ID='RSNinfo', info=(
            b'\x01\x00'              # RSN Version 1
            b'\x00\x0f\xac\x04'      # Group Cipher Suite: AES (CCMP)
            b'\x01\x00'              # 1 Pairwise Cipher Suite
            b'\x00\x0f\xac\x04'      # Pairwise Cipher Suite: AES (CCMP)
            b'\x01\x00'              # 1 Authentication Key Management Suite (AKM)
            b'\x00\x0f\xac\x02'      # AKM Suite: PSK
            b'\xac\x00'              # RSN Capabilities (MFP capable)
        ))
        elif self.to_craft == 'Dot11ProbeReq':
            iprint("This packet type requires only two addresses to be specified, overriding addr3 to addr2...")
            broadcast = cinput("I assume you want to broadcast (visible to all devices) (enter/n)")
            self.addr1 = self.addr1 if broadcast == "n" else "FF:FF:FF:FF:FF:FF"
            dot11 = Dot11(type=0, subtype=4, addr1=self.addr1, addr2=self.addr2, addr3=self.addr2)
            probe_req = Dot11ProbeReq()
            ssid = Dot11Elt(ID="SSID", info=self.ssid, len=len(self.ssid))
            packet = radiotap / dot11 / probe_req / ssid
        elif self.to_craft == 'Dot11ProbeResp':
            iprint("This packet type requires only two addresses to be specified, overriding addr3 to addr2...")
            dot11 = Dot11(type=0, subtype=5, addr1=self.addr1, addr2=self.addr2, addr3=self.addr2)
            cap = cinput("Encryption code (default: ESS+privacy)") or "ESS+privacy"
            rates = cinput("Enter byte sequence of supported rates (default: \\x82\\x84\\x8b\\x96\\x0c\\x12\\x18\\x24)") or b'\x82\x84\x8b\x96\x0c\x12\x18\x24'
            probe_resp = Dot11ProbeResp(timestamp=int(time.time()*1000000), beacon_interval=0x0064, cap=cap)
            ssid_elt = Dot11Elt(ID='SSID', info=self.ssid.encode())
            rates_elt = Dot11Elt(ID='Rates', info=rates)
            dsset_elt = Dot11Elt(ID='DSset', info=chr(1))
            packet = radiotap / dot11 / probe_resp / ssid_elt / rates_elt / dsset_elt
        elif self.to_craft == 'Dot11AssoReq':
            # 1 - AP Mac
            # 2 - Client Mac
            # 3 - AP Mac
            iprint("This packet type requires only two addresses to be specified, where addr1/3 is AP Mac, and addr2 is Client Mac, overriding addr3 to addr1...")
            ssid = Dot11Elt(ID="SSID", info=self.ssid, len=len(str(self.ssid)))
            packet = radiotap / Dot11(type=0, subtype=0, addr1=self.addr1, addr2=self.addr2, addr3=self.addr1) / Dot11AssoReq(cap=0x1100, listen_interval=0x00a) / ssid
        elif self.to_craft == 'Dot11AssoResp':
            # 1 - Client Mac
            # 2 - AP Mac
            # 3 - AP Mac
            iprint("This packet type requires only two addresses to be specified, where addr2/3 is AP Mac, and addr1 is Client Mac, overriding addr3 to addr2...")
            status = cinput("Association Response status code (default: 0)") or 0
            aid = cinput("Association ID (AID) (default: 1)") or 1
            cap = cinput("Encryption code (default: ESS+privacy)") or "ESS+privacy"
            dot11 = Dot11(addr1=self.addr1, addr2=self.addr2, addr3=self.addr2)
            packet = radiotap / dot11 / Dot11AssoResp(cap=cap, status=int(status), AID=int(aid))
        elif self.to_craft == 'Dot11Auth':
            # 1 - AP Mac
            # 2 - Client Mac
            # 3 - AP Mac
            iprint("This packet type requires only two addresses to be specified, where addr1/3 is AP Mac, and addr2 is Client Mac, overriding addr3 to addr1...")
            status = cinput("Auth status code (default: 0)") or 0
            algo = cinput("Algo (Authentication algorithm) code (default: 0)") or 0
            seqnum = cinput("Seqnum (which auth message is this one) code (default: 1)") or 1
            packet = radiotap / Dot11(type=0, subtype=11, addr1=self.addr1, addr2=self.addr2, addr3=self.addr1) / Dot11Auth(algo=int(algo), seqnum=int(seqnum), status=int(status))
        elif self.to_craft == 'Dot11Deauth':
            # 1 - Client Mac
            # 2 - AP Mac
            # 3 - AP Mac
            iprint("This packet type requires only two addresses to be specified, where addr2/3 is AP Mac, and addr1 is Client Mac, overriding addr3 to addr2...")
            dot11 = Dot11(addr1=self.addr1, addr2=self.addr2, addr3=self.addr2)
            deauth = Dot11Deauth(reason=7)
            packet = radiotap / dot11 / deauth
        elif self.to_craft == 'Dot11Disas':
            # 1 - Client Mac
            # 2 - AP Mac
            # 3 - AP Mac
            reason = cinput("Disassociation reason code (default: 7)") or 7
            dot11 = Dot11(addr1=self.addr1, addr2=self.addr2, addr3=self.addr2)
            disas = Dot11Disas(reason=int(reason))
            packet = radiotap / dot11 / disas
        elif self.to_craft == 'Dot11QoS':
            TID = cinput("Traffic Identifier (TID) code (default: 0)") or 0
            EOSP = cinput("End of Service Period (EOSP) code (default: 0)") or 0
            Acks = cinput("Acknowledge policy (default: 0)") or 0
            dot11 = Dot11(addr1=self.addr1, addr2=self.addr2, addr3=self.addr3)
            qos = Dot11QoS(TID=int(TID), EOSP=int(EOSP), Ack_Policy=int(Acks))
            packet = radiotap / dot11 / qos
        elif self.to_craft == 'Dot11RTS':
            # 1 - Receiver address
            # 2 - Transmitter address
            packet = radiotap / Dot11(type=1, subtype=11, addr1=self.addr1, addr2=self.addr2)
        elif self.to_craft == 'Dot11CTS':
            # 1 - Receiver address
            packet = radiotap / Dot11(type=1, subtype=12, addr1=self.addr1)
        elif self.to_craft == 'Dot11Ack':
            # 1 - Receiver address
            packet = radiotap / Dot11(type=1, subtype=13, addr1=self.addr1)
        elif self.to_craft == 'Dot11CFEnd':
            # 1 - Receiver address
            # 2 - BSSID
            packet = radiotap / Dot11(type=1, subtype=14, addr1=self.addr1, addr2=self.addr2)
        elif self.to_craft == 'EAPOL':
            eapol_type = cinput("Enter EAPOL packet type (Start, Logoff, Key, ASF-Alert) (default: Start)") or "Start"
            packet = self.create_eapol_packet(eapol_type)

        return packet

    def create_eapol_packet(self, eapol_type):
        eapol_type = eapol_type.lower()
        
        if eapol_type == 'start':
            iprint("Constructing an EAPOL-Start packet...")
            packet = EAPOL(type=1)
        elif eapol_type == 'logoff':
            iprint("Constructing an EAPOL-Logoff packet...")
            packet = EAPOL(type=2)
        elif eapol_type == 'key':
            iprint("Constructing an EAPOL-Key packet...")
            key_descriptor_type = cinput("Enter Key Descriptor Type (default: 2)") or 2
            key_info = cinput("Enter Key Information (default: 0x8a02)") or 0x8a02
            key_length = cinput("Enter Key Length (default: 16)") or 16
            key_replay_counter = cinput("Enter Key Replay Counter (default: 1)") or 1
            key_nonce = cinput("Enter Key Nonce (default: '00' * 32)") or '00' * 32
            key_iv = cinput("Enter Key IV (default: '00' * 16)") or '00' * 16
            key_rsc = cinput("Enter Key RSC (default: '00' * 8)") or '00' * 8
            key_id = cinput("Enter Key ID (default: '00' * 8)") or '00' * 8
            key_mic = cinput("Enter Key MIC (default: '00' * 16)") or '00' * 16
            key_data_length = cinput("Enter Key Data Length (default: 0)") or 0
            key_data = cinput("Enter Key Data (default: '')") or ''
            
            packet = EAPOL(type=3) / EAPOL.Key(
                key_descriptor_type=int(key_descriptor_type),
                key_info=int(key_info, 16),
                key_length=int(key_length),
                replay_counter=int(key_replay_counter),
                nonce=bytes.fromhex(key_nonce),
                key_iv=bytes.fromhex(key_iv),
                key_rsc=bytes.fromhex(key_rsc),
                key_id=bytes.fromhex(key_id),
                mic=bytes.fromhex(key_mic),
                wpa_key_length=int(key_data_length),
                wpa_key=bytes.fromhex(key_data)
            )
        elif eapol_type == 'asf-alert':
            iprint("Constructing an EAPOL-Encapsulated-ASF-Alert packet...")
            asf_data = cinput("Enter ASF Data (default: '00' * 32)") or '00' * 32
            packet = EAPOL(type=0) / Raw(bytes.fromhex(asf_data))
        else:
            wprint(f"Unknown EAPOL packet type: {eapol_type}. Defaulting to EAPOL-Start.")
            packet = EAPOL(type=1)
        
        return packet

    def send_packet(self, packet):
        sent = 0
        while not self.global_event.is_set():
            if self.count >= sent and not self.loop:
                self.global_event.set()
                break
            try:
                for interface in self.interface:
                    sendp(packet, iface=interface, count=1, inter=self.interval, verbose=0)
                    sent += 1
            except OSError:
                wprint("The network interface is down, but who cares? Trying to put it up...")

    def start_sending(self):
        self.global_event.clear()
        packet = self.craft_packet()
        if packet:
            iprint(f"Sending {self.count if not self.loop else 'infinite'} {self.to_craft} packet(s)...")
            iprint(f"Packet format: ADDR1 {self.addr1} / ADDR2 {self.addr2} / ADDR3 {self.addr3} / SSID {self.ssid}")
            threads = []
            for _ in range(self.threads):
                t = threading.Thread(target=self.send_packet, args=(packet,))
                t.start()
                threads.append(t)
            try:
                while not self.global_event.is_set():
                    pass
            except KeyboardInterrupt:
                self.global_event.set()
                iprint("Stopping all threads...")
            for t in threads:
                t.join()
        else:
            wprint("Failed to craft packet.")

if __name__ == "__main__":
    interface = "wlan0"
    to_craft = cinput("Specify packet to craft")
    addr1 = cinput("Enter address1 (or leave empty for random)") or random_mac()
    addr2 = cinput("Enter address2 (or leave empty for random)") or random_mac()
    addr3 = cinput("Enter address3 (or leave empty for random)") or random_mac()
    ssid = cinput("Enter SSID (optional)") or "Freeway"
    threads = 2
    count = 10
    interval = 0.1

    crafting_table = CraftingTable(interface, to_craft, addr1, addr2, addr3, ssid, threads, count, interval, loop=True)
    crafting_table.start_sending()




"""
    Simple script that downloads github files
"""

import os
import requests
import base64
import sys
import time

try:
    from FreewayTools.colors import cprint, iprint, wprint, cinput, ColorCodes

except ModuleNotFoundError:
    from colors import cprint, iprint, wprint, cinput, ColorCodes

def download_folder_from_github(owner, repo, path, local_dir, token=None):
    if not os.path.exists(local_dir):
        os.makedirs(local_dir)

    url = f"https://api.github.com/repos/{owner}/{repo}/contents/{path}"
    headers = {
        "Accept": "application/vnd.github.v3+json",
    }
    if token:
        headers["Authorization"] = f"Bearer {token}"

    try:
        response = requests.get(url, headers=headers)
        if response.status_code == 200:
            files = response.json()
            for file in files:
                if file['type'] == 'file':
                    download_url = file['download_url']
                    file_response = requests.get(download_url)
                    local_file_path = os.path.join(local_dir, file['name'])
                    with open(local_file_path, 'wb') as local_file:
                        local_file.write(file_response.content)
                    cprint(f"Downloaded: {file['name']}")
                else:
                    # If there are subdirectories, call the function recursively
                    download_folder_from_github(owner, repo, file['path'], os.path.join(local_dir, file['name']), token)
        else:
            wprint(f"Failed to get contents of directory: STATUS_CODE {response.status_code}")
    except Exception as e:
        wprint(f"Failed to download {path}! You may need to download manually or check for your internet connection.")
        time.sleep(3)


"""
    IEEE 802.11 Packet Monitor using curses module as display,
    + captures PMKIDs & Handshakes + catches most clients
    + checks manufacturer + filter options
"""
import curses
import time
from datetime import datetime, timedelta
from scapy.all import *
from scapy.layers.dot11 import Dot11Beacon, Dot11Elt, Dot11ProbeReq, Dot11ProbeResp, Dot11AssoReq, Dot11Auth, Dot11, Dot11Deauth, RadioTap, Dot11AssoResp, Dot11Disas, Dot11QoS
from scapy.layers.eap import EAPOL
import logging

try:
    from FreewayTools.checkmac import check_manufacturer
except ModuleNotFoundError:
    from checkmac import check_manufacturer

log_dir = "/usr/local/share/3way/"

if not os.path.exists(log_dir):
    os.makedirs(log_dir)

logging.basicConfig(filename=os.path.join(log_dir, 'debug.log'), level=logging.DEBUG)

def init_colors():
    curses.start_color()
    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_BLUE, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
    curses.init_pair(4, curses.COLOR_CYAN, curses.COLOR_BLACK)
    curses.init_pair(5, curses.COLOR_RED, curses.COLOR_BLACK)
    curses.init_pair(6, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    curses.init_pair(7, curses.COLOR_WHITE, curses.COLOR_BLACK)
    curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_WHITE) 

def get_signal_strength(packet):
    """
    Extracts signal strength in dBm from the packet if available
    """
    if packet.haslayer(RadioTap):
        radiotap_fields = packet[RadioTap].fields
        if 'dBm_AntSignal' in radiotap_fields:
            return radiotap_fields['dBm_AntSignal']
    return None

def save_hash_file(path, content):
    try:
        with open(path + ".hash", "w") as f:
            f.write(content)
    except Exception as e:
        logging.debug(str(e))

class PacketHandler:
    def __init__(self, aps, clients, parms, script_dir):
        self.script_dir = script_dir
        self.APs = aps
        self.Clients = clients
        self.parms = parms
        self.handshake_dir = os.path.join(self.script_dir, "caps/handshakes")
        self.pmkids_dir = os.path.join(self.script_dir, "caps/pmkids")
        os.makedirs(self.handshake_dir, exist_ok=True)
        os.makedirs(self.pmkids_dir, exist_ok=True)
        self.handshakes = {}
        self.pmkids = {}

    def determine_encryption(self, packet):
        """Determines encryption type of the AP"""
        encryption = "Open"
        if packet.haslayer(Dot11Elt):
            rsn_found = packet.getlayer(Dot11Elt, ID=48)
            wpa_found = packet.getlayer(Dot11Elt, ID=221)
            
            if rsn_found:
                encryption = self.parse_rsn_element(rsn_found)

            elif wpa_found and b'\x00P\xf2\x01\x01\x00' in wpa_found.info:
                encryption = "WPA"

        return encryption

    def parse_rsn_element(self, rsn_elt):
        """Determines encryption type included in RSN"""
        encryption = "WPA2"

        if len(rsn_elt.info) > 2:
            asc = int.from_bytes(rsn_elt.info[10:12], byteorder='little')

            if asc > 0:
                akm_offset = 12
                for _ in range(asc):
                    akm_suite = rsn_elt.info[akm_offset:akm_offset+4]
                    if akm_suite in [b'\x00\x0F\xAC\x02', b'\x00\x0F\xAC\x05']:
                        encryption += " (Personal)"
                    elif akm_suite in [b'\x00\x0F\xAC\x01', b'\x00\x0F\xAC\x05']:
                        encryption = "WPA2 Enterprise"
                    elif akm_suite in [b'\x00\x0F\xAC\x08', b'\x00\x0F\xAC\x09', b'\x00\x0F\xAC\x12']:
                        encryption = "WPA3 Enterprise"
                    elif akm_suite == b'\x00\x0F\xAC\x04':
                        encryption = "WPA3"

                    akm_offset += 4

        return encryption

    def get_packet_type(self, packet):
        """Determines the type of a WiFi packet."""
        if packet.haslayer(Dot11Beacon):
            return "Beacon"
        elif packet.haslayer(Dot11ProbeReq):
            return "Probe Request"
        elif packet.haslayer(Dot11ProbeResp):
            return "Probe Response"
        elif packet.haslayer(Dot11AssoReq) or packet.haslayer(Dot11AssoResp):
            return "Association"
        elif packet.haslayer(Dot11Auth):
            return "Authentication"
        elif packet.haslayer(Dot11Deauth):
            return "Deauthentication"
        elif packet.haslayer(Dot11Disas):
            return "Disassociation"
        elif packet.haslayer(Dot11QoS):
            return "QoS Data"
        else:
            if packet.type == 1:
                subtype = packet.subtype
                if subtype == 0x09:
                    return "Block Ack Req"
                elif subtype == 0x0b:
                    return "Block Ack"
                elif subtype == 0x0B:
                    return "RTS"
                elif subtype == 0x0C:
                    return "CTS"
            return "Other"

    def packet_breathalyzer(self, ap_mac, client_mac, types):
        """
            The author of this function needs a breathalyzer himself (me) 
            *Sorry for such a mess
        """
        if types in ["Association", "Authentication", "QoS Data", "RTS", "Block Ack"]:
            if client_mac in self.APs or ap_mac in self.APs:
                self.APs[ap_mac if ap_mac in self.APs else client_mac]["Clients"].add(client_mac if client_mac not in self.APs else ap_mac)
                logging.debug(f"PB-{types}: Current client list for {ap_mac}: {self.APs[ap_mac if ap_mac in self.APs else client_mac]['Clients']}")
            if client_mac in self.Clients:
                self.Clients[client_mac]["AP_MAC"] = ap_mac
            elif ap_mac in self.Clients:
                self.Clients[ap_mac]["AP_MAC"] = client_mac

        elif types in ["Deauthentication", "Disassociation"]:
            if client_mac in self.APs or ap_mac in self.APs:
                if client_mac in self.APs[ap_mac if ap_mac in self.APs else client_mac]["Clients"] or ap_mac in self.APs[ap_mac if ap_mac in self.APs else client_mac]["Clients"]:
                    self.APs[ap_mac if ap_mac in self.APs else client_mac]["Clients"].remove(client_mac if client_mac not in self.APs else ap_mac)

    def extract_channel(self, packet):
        channel = "N/A"
        if packet.haslayer(RadioTap):
            try:
                channel = ord(packet[Dot11Elt:3].info)
            except Exception as e:
                logging.debug(f"In function extract_channel: {type(e).__name__}: {str(e)}")
        return channel

    def handle_dot11_beacon(self, packet, timestamp):
        mac_address = packet[Dot11].addr2
        channel = self.extract_channel(packet)
        ssid = packet[Dot11Elt].info.decode('utf-8', 'ignore') if packet[Dot11Elt].info else "Hidden SSID"
        enc = self.determine_encryption(packet)
        signal_strength = get_signal_strength(packet)
        manufacturer = check_manufacturer(mac_address)

        if mac_address not in self.APs:
            self.APs[mac_address] = {"SSID": ssid, "Clients": set(), "PMKIDs": set(), "Handshakes": set(), "Last Beacon": timestamp, "ENC": enc, "Signal": signal_strength, "Channel": channel, "Manufacturer": manufacturer}
        else:
            self.APs[mac_address]["ENC"] = enc
            self.APs[mac_address]["Signal"] = signal_strength
            self.APs[mac_address]["Channel"] = channel
            self.APs[mac_address]["Manufacturer"] = manufacturer
        self._handle_client_activity(packet, timestamp, update_ap=True)

    def handle_dot11_probe_req(self, packet, timestamp):
        mac_address = packet[Dot11].addr2

        signal = get_signal_strength(packet)

        if packet.haslayer(Dot11Elt):
            for element in packet[Dot11Elt]:
                if element.ID == 0 and element.info: # ID 0 == SSID
                    if mac_address in self.Clients:
                        ssid = element.info.decode('utf-8', 'ignore')
                        self.Clients[mac_address]["Name"].add(ssid)
                        logging.debug(f"Current probes list for {mac_address}: {self.Clients[mac_address]['Name']}")
                        return
        self._handle_client_activity(packet, timestamp, update_ap=True)

    def handle_dot11_probe_resp(self, packet, timestamp):
        ap_mac = packet[Dot11].addr2
        enc = self.determine_encryption(packet)
        if ap_mac in self.APs:
            self.APs[ap_mac]["ENC"] = enc
        self._handle_client_activity(packet, timestamp, update_ap=True)

    def handle_dot11_asso_req(self, packet, timestamp):
        self._handle_client_activity(packet, timestamp, update_ap=True)

    def handle_dot11_auth(self, packet, timestamp):
        self._handle_client_activity(packet, timestamp, update_ap=True)

    def handle_eapol_frame(self, packet, timestamp):
        """
            Capture WPA 2/3 handshakes and PMKIDs for all sessions.
        """

        src_mac = packet[Dot11].addr2
        dst_mac = packet[Dot11].addr1
        session_id = f"{src_mac}-{dst_mac}"
        session_id_reversed = f"{dst_mac}-{src_mac}"
        logging.debug(f"{timestamp}--Detected EAPOL Packet! Logging session-id: {session_id}")

        if session_id not in self.handshakes or session_id not in self.pmkids:
            self.handshakes[session_id] = {'to_frames': 0, 'from_frames': 0, 'packets': []}
            self.pmkids[session_id] = {'num_frame': 0, 'first_eapol_frame': None, 'pmkid': None, 'mac_ap': None, 'mac_cl': None, 'packets': []} 

        # PMKIDs
        if session_id_reversed in self.pmkids:
            temp_data = self.pmkids[session_id]
            del self.pmkids[session_id]
            self.pmkids[session_id_reversed]['num_frame'] = temp_data['num_frame']
            session_id = session_id_reversed

        self.pmkids[session_id]['num_frame'] += 1

        if self.pmkids[session_id]['num_frame'] == 1:
            self.pmkids[session_id]['first_eapol_frame'] = bytes(packet[EAPOL])
            self.pmkids[session_id]['pmkid'] = self.pmkids[session_id]['first_eapol_frame'][-16:].hex()
            self.pmkids[session_id]['mac_ap'] = packet.addr2
            self.pmkids[session_id]['mac_cl'] = packet.addr1
            logging.debug(f"Detected 1st EAPOL PMKID packet for {self.pmkids[session_id]}")

        if self.pmkids[session_id]['num_frame'] == 2:
            if not src_mac in self.APs:
                logging.debug(f"Detected second EAPOL PMKID packet but we didnt catch the SSID yet! Session: {self.pmkids[session_id]}")
                return
            self.APs[src_mac]["PMKIDs"].add(timestamp)
            logging.debug("\n1st EAPoL Frame:   \n" + str(self.pmkids[session_id]['first_eapol_frame']) + "\n")
            logging.debug("Possible PMKID:        ", self.pmkids[session_id]['pmkid'])
            ssid_hex = bytes(self.APs[src_mac]["SSID"], 'utf-8').hex()
            logging.debug("SSID:                  ", self.APs[src_mac]["SSID"])
            mac_ap_formatted = self.pmkids[session_id]['mac_ap'].replace(":", "").lower()
            mac_cl_formatted = self.pmkids[session_id]['mac_cl'].replace(":", "").lower()
            logging.debug("MAC AP:                ", mac_ap_formatted)
            logging.debug("MAC Client:            ", mac_cl_formatted)
            logging.debug("\nEncoded PMKID compatible with Hashcat hc22000:")
            hash_line = f"{self.pmkids[session_id]['pmkid']}*{mac_ap_formatted}*{mac_cl_formatted}*{ssid_hex}"
            logging.debug(hash_line)
            save_hash_file(os.path.join(self.pmkids_dir, "pmkid_" + timestamp), hash_line)

        # Handshakes
        to_ds = packet.FCfield & 0x1 != 0
        from_ds = packet.FCfield & 0x2 != 0

        if to_ds and not from_ds:
            self.handshakes[session_id]['to_frames'] += 1
        elif not to_ds and from_ds:
            self.handshakes[session_id]['from_frames'] += 1

        self.handshakes[session_id]['packets'].append(packet)

        filename = os.path.join(self.handshake_dir, f"{session_id}_handshake.pcap")
        dump = PcapWriter(filename, append=True, sync=True)
        dump.write(packet)

        if self.handshakes[session_id]['to_frames'] >= 2 and self.handshakes[session_id]['from_frames'] >= 2:
            if self.APs[src_mac]:
                self.APs[src_mac]["Handshakes"].add(timestamp)
            logging.debug(f"Captured complete WPA handshake for session {session_id}")
            self.handshakes[session_id] = {'to_frames': 0, 'from_frames': 0, 'packets': []}
        self._handle_client_activity(packet, timestamp, update_ap=True)

    def _handle_client_activity(self, packet, timestamp, update_ap=False):
        ptype = self.get_packet_type(packet)
        ap_mac = packet[Dot11].addr2
        mac_address = packet[Dot11].addr1 if ptype not in ["Probe Request", "CTS", "RTS"] else packet[Dot11].addr2

        if ap_mac in self.APs:
            for mac, ap_info in self.APs.items():
                if mac_address in ap_info['Clients']:
                    self.APs[mac]['Clients'].remove(mac_address)
                    self.APs[ap_mac]['Clients'].add(mac_address)

        self.packet_breathalyzer(ap_mac, mac_address, ptype)

        signal = get_signal_strength(packet)

        if mac_address in self.APs and mac_address in self.Clients:
            del self.Clients[mac_address]
        elif ap_mac in self.APs and ap_mac in self.Clients:
            del self.Clients[ap_mac]

        if update_ap and ap_mac in self.APs and mac_address in self.Clients:
            ssid = self.APs[ap_mac]["SSID"]
            if ssid not in self.Clients[mac_address]["Name"]:
                self.Clients[mac_address]["Name"].add(ssid)
        else:
            if mac_address not in ["ff:ff:ff:ff:ff:ff", None]:
                if mac_address not in self.Clients and mac_address not in self.APs:
                    self.Clients[mac_address] = {"Name": set(), "AP_MAC": "Not Associated", "Last Beacon": timestamp, "Signal": signal, "Manufacturer": check_manufacturer(mac_address)}
                    for mac, ap_info in self.APs.items():
                        if mac_address in ap_info['Clients']:
                            self.Clients[mac_address]['AP_MAC'] = mac

class Monitor:
    def __init__(self, interface, script_dir=None, parms=None, filters=None):
        self.script_dir = script_dir
        self.interface = interface
        self.parms = parms or {}
        self.filters = filters or {}
        self.save = True if "f" in self.filters else False
        self.caps_dir = os.path.join(self.script_dir, "caps")
        self.save_dir = os.path.join(self.script_dir, "caps/saved")
        self.session = self.new_session_number()
        os.makedirs(self.caps_dir, exist_ok=True)
        os.makedirs(self.save_dir, exist_ok=True)
        self.APs = {}
        self.Clients = {}
        self.run_time = time.time()
        self.packet_handler = PacketHandler(self.APs, self.Clients, self.parms, self.script_dir)

    def new_session_number(self):
        session_time = datetime.now().strftime('%m%d_%H%M%S')
        random_string = ''.join(random.choices(string.ascii_letters + string.digits, k=random.randint(8, 12)))
        unique_id = f"{session_time}_{random_string}"
        return unique_id

    def process_packet(self, packet):
        timestamp = datetime.now().strftime('%m-%d %H:%M:%S')
        if self.save:
            PcapWriter(f"{self.save_dir}/{self.session}", append=True, sync=True).write(packet)
        if EAPOL in packet or packet.haslayer(EAPOL):
            self.packet_handler.handle_eapol_frame(packet, timestamp)
        elif packet.haslayer(Dot11):
            if packet.haslayer(Dot11Beacon):
                self.packet_handler.handle_dot11_beacon(packet, timestamp)
            elif packet.haslayer(Dot11ProbeReq):
                self.packet_handler.handle_dot11_probe_req(packet, timestamp)
            elif packet.haslayer(Dot11ProbeResp):
                self.packet_handler.handle_dot11_probe_resp(packet, timestamp)
            elif packet.haslayer(Dot11AssoReq):
                self.packet_handler.handle_dot11_asso_req(packet, timestamp)
            elif packet.haslayer(Dot11Auth):
                self.packet_handler.handle_dot11_auth(packet, timestamp)
            else:
                self.packet_handler._handle_client_activity(packet, timestamp, update_ap=True)
    
    def determine_columns_count(self):
        column_names = ["SSID", "MAC", "CLIENTS", "ENC", "PMKIDs", "HANDSHAKES", "LAST BEACON", "SIGNAL"]
        if 'channel' in self.parms:
            column_names.append("CH.")
        if 'manu' in self.parms:
            column_names.append("MANUFACTURER")

        return len(column_names)

    def get_dynamic_widths(self, stdscr, padding=2):
        _, width = stdscr.getmaxyx() 
        num_columns = self.determine_columns_count()
        available_width = width - (padding * (num_columns - 1))
        ssid_width = int(available_width * 0.17)
        mac_width = int(available_width * 0.2) if num_columns == 8 else int(available_width * 0.17)
        clients_width = int(available_width * 0.1) if num_columns == 8 else int(available_width * 0.08)
        enc_width = int(available_width * 0.1) if num_columns == 8 else int(available_width * 0.08)
        pmkids_width = int(available_width * 0.08) 
        handshakes_width = int(available_width * 0.12) if num_columns == 8 else int(available_width * 0.1)
        last_beacon_width = int(available_width * 0.16) if num_columns == 8 else int(available_width * 0.10)
        signal_width = int(available_width * 0.05)
        channel_width = int(available_width * 0.05) if 'channel' in self.parms else 0
        manu_width = int(available_width * 0.12) if 'manu' in self.parms else 0
        return ssid_width, mac_width, clients_width, enc_width, pmkids_width, handshakes_width, last_beacon_width, signal_width, channel_width, manu_width

    def safe_addstr(self, stdscr, y, x, text, attr=None):
        """ 
            Ensures the text can be added to the view
            basing on terminal width and height
        """
        max_y, max_x = stdscr.getmaxyx()
        if attr is None:
            attr = curses.A_NORMAL

        if y < max_y and x < max_x:
            try:
                stdscr.addstr(y, x, text, attr)
            except curses.error:
                pass
            except ValueError:
                pass

    def calc_time(self, elapsed_seconds):
        """Calculate the time passed"""
        elapsed_time = timedelta(seconds=elapsed_seconds)
        
        days, remainder = divmod(elapsed_time.total_seconds(), 86400)
        hours, remainder = divmod(remainder, 3600)
        minutes = remainder // 60

        runtime_parts = []
        if days > 0:
            runtime_parts.append(f"{int(days)}d")
        if hours > 0:
            runtime_parts.append(f"{int(hours)}h")
        if minutes > 0 or (days == 0 and hours == 0):
            runtime_parts.append(f"{int(minutes)}m")

        return runtime_parts

    def show_captured_data(self, stdscr):
        """
            Streams all the captured beacon packets onto the curses view,
            initializes and calculates banners and other styling elements
        """
        current_time = time.time()
        stdscr.clear()
        height, width = stdscr.getmaxyx()

        display_aps = True if not "c" in self.filters else False
        display_clients = True if not "a" in self.filters else False
        display_no_empty_aps = True if "e" in self.filters else False
        display_stack_ssids = True if "s" in self.filters else False
        display_no_lonely_clients = True if "n" in self.filters else False

        banner = [
            " ______                       ",
            "|__    |.--.--.--.---.-.--.--.",
            "|__    ||  |  |  |  _  |  |  |",
            "|______||________|___._|___  |",
            "                       |_____|"
        ]

        start_row_banner = 0
        start_col = 0
        
        # freeway banner setup
        for line in banner:
            start_col = max(0, (width // 2) - (len(line) // 2))
            end_col = start_col + len(line)

            part_length = len(line) // 3
            part1 = line[:part_length]
            part2 = line[part_length:2*part_length]
            part3 = line[2*part_length:]

            self.safe_addstr(stdscr, start_row_banner, start_col, part1, curses.color_pair(1) | curses.A_BOLD)
            self.safe_addstr(stdscr, start_row_banner, start_col + len(part1), part2, curses.color_pair(2) | curses.A_BOLD)
            self.safe_addstr(stdscr, start_row_banner, start_col + len(part1) + len(part2), part3, curses.color_pair(5) | curses.A_BOLD)
            
            start_row_banner += 1

        stdscr.refresh()

        elapsed_seconds = int(current_time - self.run_time)
        runtime_parts = self.calc_time(elapsed_seconds)

        runtime_str = " ".join(runtime_parts)
        rtheader = f"running for {runtime_str}"
        
        runtime_col = max(0, (width // 2) - (len(rtheader) // 2))

        self.safe_addstr(stdscr, start_row_banner, runtime_col + 23, rtheader, curses.color_pair(4) | curses.A_BOLD)
        self.row = start_row_banner + 1

        ssid_width, mac_width, clients_width, enc_width, pmkids_width, handshakes_width, last_beacon_width, signal_width, channel_width, manu_width = self.get_dynamic_widths(stdscr)
        if display_aps:
            header_elements = ["SSID", "MAC", "CLIENTS", "ENC", "PMKIDs", "HANDSHAKES", "LAST BEACON", "SIGNAL"]
            ap_header_format = " ".join(["{:<" + str(width) + "}" for width in (ssid_width, mac_width, clients_width, enc_width, pmkids_width, handshakes_width, last_beacon_width, signal_width)])

            if 'channel' in self.parms:
                header_elements.append("CH.")
                ap_header_format += " {:<" + str(channel_width) + "}"

            if 'manu' in self.parms:
                header_elements.append("MANUFACTURER")
                ap_header_format += " {:<" + str(manu_width) + "}"            

            ap_header = ap_header_format.format(*header_elements)

            self.safe_addstr(stdscr, self.row, 0, "[>]   Access Points", curses.color_pair(2))
            self.row += 1
            self.safe_addstr(stdscr, self.row + 1, 0, ap_header, curses.color_pair(3))
            self.row += 2

            self.known_ssids = set()

            # APs data
            for mac, info in self.APs.items():
                if (display_no_empty_aps and len(info['Clients']) == 0) or (display_stack_ssids and 
                info['SSID'] in self.known_ssids):
                    continue

                signal = info.get("Signal", 0)
                if signal < -50:
                    color_pair = curses.color_pair(2)
                else:
                    color_pair = curses.color_pair(1)

                ssid = (info['SSID'][:ssid_width - 3] + '...') if len(info['SSID']) > ssid_width else info['SSID']
                self.known_ssids.add(info['SSID'])
                enc = (info['ENC'][:enc_width - 3] + '...') if len(info['ENC']) > enc_width else info['ENC']
                signal_str = str(signal) + " dBm"

                ap_line = "{:<{}} {:<{}} {:<{}} {:<{}} {:<{}} {:<{}} {:<{}} {:<{}} ".format(ssid, ssid_width, mac, mac_width, str(len(info['Clients'])), clients_width, enc, enc_width, str(len(info['PMKIDs'])), pmkids_width, str(len(info['Handshakes'])), handshakes_width, info['Last Beacon'], last_beacon_width, signal_str, signal_width)
                
                if "channel" in self.parms:
                    ap_line += "{:<{}}".format(info.get("Channel", "N/A"), channel_width)
                if "manu" in self.parms:
                    ap_line += "{:<{}}".format(info.get("Manufacturer"), manu_width)

                self.safe_addstr(stdscr, self.row, 0, ap_line, color_pair)
                self.row += 1

        if display_clients:
            self.row += 1
            client_header_format = "{:<" + str(ssid_width) + "} {:<" + str(mac_width) + "} {:<" + str(mac_width) + "} {:<" + str(last_beacon_width) + "} {:<" + str(signal_width) + "} {:<" + str(manu_width) + "}"
            self.safe_addstr(stdscr, self.row, 0, "[+]   Clients", curses.color_pair(5))
            self.row += 2
            client_header = client_header_format.format("NAME (Probe)", "MAC", "AP MAC", "LAST BEACON", "SIGNAL", "MANUFACTURER")
            self.safe_addstr(stdscr, self.row, 0, client_header, curses.color_pair(3))
            self.row += 1

            # Clients data
            for mac, info in self.Clients.items():
                if (display_no_lonely_clients and len(info['AP_MAC']) < 16):
                    continue

                full_names = ', '.join(info['Name']) if 'Name' in info else " "
                if len(full_names) > ssid_width:
                    full_names = full_names[:ssid_width - 3] + '...'

                ap_mac = info['AP_MAC'] or ''
                last_beacon = info['Last Beacon']
                signal_str = str(info['Signal']) + "dbm"
                manufacturer = info['Manufacturer']
                client_line = client_header_format.format(full_names, mac, ap_mac, last_beacon, signal_str, manufacturer)
                self.safe_addstr(stdscr, self.row, 0, client_line, curses.color_pair(4))
                self.row += 1

        stdscr.refresh()

    def start_sniffing(self, stdscr):
        self.mod_curses('enter', stdscr)

        running = True
        network_down_message_displayed = False
        counter = 0
        while running:
            try:
                sniff(iface=self.interface, count=4, prn=self.process_packet, store=0, monitor=True)
                if network_down_message_displayed and counter <= 10:
                    counter += 1
                    self.safe_addstr(stdscr, self.row + 3, 0, "Network is down..", curses.color_pair(5))
                    stdscr.refresh()                    

                    if counter == 10:
                        network_down_message_displayed = False
                        counter = 0
                        
            except OSError as e:
                if e.errno == 100:
                    if not network_down_message_displayed:
                        self.safe_addstr(stdscr, self.row + 3, 0, "Network is down..", curses.color_pair(5))
                        stdscr.refresh()
                        network_down_message_displayed = True
                        time.sleep(3)
                else:
                    raise

            self.show_captured_data(stdscr)
            self.safe_addstr(stdscr, self.row + 2, 0, " Press q to stop monitoring", curses.color_pair(1))
            stdscr.refresh()

            if stdscr.getch() == ord('q'):
                running = False

        # Results Screen
        self.show_captured_data(stdscr)
        self.safe_addstr(stdscr, self.row + 4 if network_down_message_displayed else self.row + 3, 0, " Press anything to exit", curses.color_pair(1))
        stdscr.refresh()

        self.mod_curses('leave', stdscr)

    def mod_curses(self, mod, stdscr):
        if mod == "enter":
            curses.curs_set(0)
            init_colors()
            stdscr.nodelay(1)
            stdscr.clear()
            return

        stdscr.nodelay(0)
        stdscr.getch()
        curses.nocbreak()
        stdscr.keypad(False)
        curses.echo()  

if __name__ == "__main__":
    interface = "wlan0mon"
    curses.wrapper(Monitor(interface).start_sniffing)


class ColorCodes:
    RED = "\033[31m"
    GREEN = "\033[32m"
    BLUE = "\033[34m"
    CYAN = "\033[36m"
    WHITE = "\033[37m"
    RESET = "\033[0m"
    BRIGHT = "\033[1m"
    YELLOW = "\033[33m"
    MAGENTA = "\033[35m"
    LIGHT_GREEN = "\033[92m"
    LIGHT_BLUE = "\033[94m"
    LIGHT_CYAN = "\033[96m"
    LIGHT_RED = "\033[91m"
    LIGHT_MAGENTA = "\033[95m"
    LIGHT_YELLOW = "\033[93m"
    LIGHT_WHITE = "\033[97m"
    BLACK = "\033[30m"
    ORANGE = "\033[38;5;208m"
    PURPLE = "\033[38;5;93m"
    DARK_GRAY = "\033[38;5;238m"
    LIGHT_GRAY = "\033[38;5;245m"
    PINK = "\033[38;5;213m"
    BROWN = "\033[38;5;130m"
    BLINK = "\033[5m"

def cprint(string, color=ColorCodes.BLUE):
    print(f'{ColorCodes.RESET}{ColorCodes.CYAN}[+]{ColorCodes.RESET}' + ' ' + f'{ColorCodes.BRIGHT}{color}{string}{ColorCodes.RESET}')

def wprint(string):
    print(f'{ColorCodes.WHITE}[+]{ColorCodes.RESET}' + ' ' + f'{ColorCodes.RED}{string}{ColorCodes.RESET}')

def iprint(string):
    print(f'{ColorCodes.BRIGHT}{ColorCodes.GREEN}[{ColorCodes.RED}3WAY{ColorCodes.GREEN}]{ColorCodes.RESET}' + ' ' + f'{ColorCodes.WHITE}{ColorCodes.BRIGHT}{string}{ColorCodes.BRIGHT}{ColorCodes.RESET} ')

def oneline(string, sys, color=ColorCodes.GREEN):
    sys.stdout.write(f'\r{ColorCodes.GREEN}[{ColorCodes.RED}3WAY{ColorCodes.GREEN}]{ColorCodes.RESET}' + ' ' + f"{color}{ColorCodes.BRIGHT}{string}{ColorCodes.RESET}")
    sys.stdout.flush()

def cinput(string, color=ColorCodes.CYAN, b=False):
    r = input(f'{ColorCodes.GREEN}[>]{ColorCodes.RESET}' + ' ' + f'{color}{ColorCodes.BRIGHT if b else color}{string}:{ColorCodes.RESET} ')
    return r


"""
    This file installs Freeway - making it available directly from the command line:
    'sudo Freeway'

    To uninstall Freeway run:
    'sudo pip uninstall 3way'
"""

from setuptools import setup, find_packages
from setuptools.command.install import install
import os
import shutil

class PostInstallCommand(install):
    """Post-installation for installation."""
    def run(self):
        install.run(self)
        main_dir = '/usr/local/share/3way'
        if not os.path.exists(main_dir):
            os.makedirs(main_dir)
        source = 'FreewayTools/lists/ssid_list.txt'
        destination = os.path.join(main_dir, "lists")
        if not os.path.exists(destination):
            os.makedirs(destination)
        if not os.path.exists(os.path.join(destination, "ssid_list.txt")):
            shutil.copy(source, os.path.join(destination, 'ssid_list.txt'))
        source = 'templates'
        destination = os.path.join(main_dir, source)
        if not os.path.exists(destination):
            os.makedirs(destination)
        for template in ["google", "Valentines", "mrhacker", "mcd"]:
            if not os.path.exists(os.path.join(destination, template)):
                shutil.copytree(f"{source}/{template}", f"{destination}/{template}")

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name='3way',
    version='1.3.0',
    author='FLOCK4H',
    url='https://github.com/FLOCK4H/Freeway',
    description='Freeway for network pentesting',
    long_description=long_description,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(),
    install_requires=["scapy", "rich"],

    scripts=['Freeway'],
    cmdclass={
        'install': PostInstallCommand,
    }
)



#!/usr/bin/env python3
from netaddr import EUI
from netaddr.core import NotRegisteredError
import argparse

def check_manufacturer(mac_address):
    try:
        mac = EUI(mac_address)
        manufacturer = mac.oui.registration().org
        return manufacturer
    except NotRegisteredError:
        return "Unknown/Random MAC"
    except Exception as e:
        return f"Error: {e}"

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Get the manufacturer of a given MAC address.')
    parser.add_argument('-m', '--mac', type=str, required=True, help='The MAC address to look up')
    args = parser.parse_args()
    manufacturer = check_manufacturer(args.mac)
    print(f'\033[34mManufacturer: {manufacturer}')